/** @file Device.h
	Device class definition and many useful constants and helper functions
*/

#pragma once
#ifndef __DEVICE_H
#define __DEVICE_H

#include <vector>
#include "ForwardDef.h"
#include "Interfaces.h"
#include "common.h"
#include "WMPacket.h"
#include "Globals.h"
#include "Parameter.h"
#include "Ports.h"
#include "IOContext.h"
#include "Context.h"
#include "ContextOverride.h"
#include "DeviceVersioning.h"

const int SOCKET_SIZE_X = 8;
const int SOCKET_SIZE_Y = 8;
const int PSOCKET_SIZE_X = 4;
const int PSOCKET_SIZE_Y = 6;

const int DEV_NAME_LEN = 24;

struct InputValidationStruc;

/** @name Parameter Reference Macros
	These are defined indirectly so that the same parameter names can be used
	both inside of the device itself, and inside it's custom dialog box, if any.
	@relates Device
*/
///@{
#define ParmIRef(x)		(params.getOwner()->GetIRef(x))
#define ParmFRef(x)		(params.getOwner()->GetFRef(x))
#define ParmBRef(x)		(params.getOwner()->GetBRef(x))
///@}

typedef UINT DEVID;

/**
	Devices are the basic building block that represent action on your data within World Machine.

	A Device takes zero or more inputs, performs some operation,
	and produces zero or more data packets as ouptut. The type of data, and operation, is completely open. 

	Extend the Device class to implement a new terrain action device.
	Typically, you would derive a new Device from one of the subclasses 
	defined in Filter.h or IODevices.h such as Generator, Filter, Combiner, etc.

	@ingroup devices
*/
class Device //: public IDevice
{
public:
	enum DEVICEPORT { INPUT, OUTPUT, PARAMIN, MASK, COORD, DEFAULTOUT, DEFAULTIN, NULLPORT = -1 };

	Device(void);
	virtual ~Device(void);

/// 


/// Creates an independent copy of this device. Clones will be an exact copy of an existing device (including device ID), and 
/// can/should share ownership of internal datastructures
	Device *Clone();		

	void Destroy();			///< Delete the device using its internally defined deletion operator

	friend class GraphSort;
	friend class DeviceWorldBuilder;
	friend class DeviceWorld;
	friend class GenericDlg;
	friend class BuildEngine;
	friend class PresetManager;

	/// Assign Device to DeviceWorld
	void SetWorld(DeviceWorld *world) { worldptr = world; }; 
	/// Retrieve DeviceWorld
	DeviceWorld *GetWorld() const { return worldptr; };	

	/// Query for a pointer to an advanced interface for this device.
	/// If the device doesn't support the requested operation, it will return nullptr
	virtual WorldMachine::AdvancedDeviceInterface* QueryInterface(WorldMachine::ADVANCED_INTERFACE_TYPE queryInt);

///@name Serialization Functions
/// Uses C++ streams for IO.
//@{
	virtual bool Load(std::istream &in, LoadContext &iocontext) =0;
	virtual bool Save(std::ostream &out, SaveContext &iocontext) =0;
//@}

/// @name UI functions
/// All of these are used internally by the World Machine UI.
	//@{
	virtual dword Color() { return 0x07f7f7f; }; 
	virtual int GetWidth() const;
	virtual int GetHeight() const;
	virtual Rect GetDeviceRect() const;
	virtual Rect4F GetDeviceRect4F() const;

	virtual char *GetDescription() =0 { return "This part can never be instanced";};
	virtual void  SetName(const char *name);
	virtual char *GetName();
	virtual char *GetTypeName() =0 { return "Unknown Device"; };

	virtual char *GetInputName(int slot);
	virtual char *GetOutputName(int slot);
	char *GetParamInputName(int slot);		// Non-overridable since it directly interfaces with parameterset

	/// Get location in Device workview
	Coord GetLoc() const { return loc; };
	/// Set UI location
	void SetLoc(Coord &newloc) { loc = newloc; }; 
	int	FindPort(Coord loc, int *io, int showstate); /// Discover the port present at a given location
	Coord GetPortLoc(int slot, int io);	///< Retrieve coordinates of a given port, relative to device origin
	Coord GetPortLoc(Port *target_port, int io = -1);	///< Retrieve coordinates of a pointer-to-Port
	//@}

	/// Get Device nametag
	char *GetStaticTag() { return lifeptrs.nametag; };	
	virtual int GetDeviceFamily() = 0;	///< Get Device family (used both for UI and crude RTTI)
	virtual void AssignNewDeviceID();	///< Assigns this device a new UID, unique to its parent world. No effect if no parent world assigned.
	/// Get Device's UID.
	int GetDeviceID() const { return device_uid; }; 

/// Layout View functions (optional!)
	virtual LayoutUIClient *GetLayoutUI() { return NULL; };

/** @name Device Origin functions
 Some devices support the concept of a device origin, where some effect should be centered. If your device supports this,
 hasOrigin() should return true and the following functions will have meaning.
*/
	//@{
	/// Return device origin, modified by the coordinate input, if any
	CoordF getOrigin();						
	CoordF getOrigin(BuildContext &context);
	/// return unmodified device origin
	CoordF getOriginRaw() { return origin; };
	void   setOrigin(CoordF &o) { origin = o; };
	//@}


/** @name World Context operations
	These functions provide convenient access to the device build results that are actually properly stored in the DeviceWorld. 
	@deprecated These functions will eventually be removed as they more properly should reside in their deviceworld.
*/
	//@{
	virtual PacketPointer	getDefaultInput(PSTATE preview);
	virtual PacketPointer	getDefaultOutput(PSTATE preview);
	virtual Device *		getDisplayDevice();
	virtual PacketPointer	getFinishedPacket(PSTATE state, int slot = 0);
	ScopedPacketAccess		viewInput(int slot, PSTATE preview);
	//@}

/// @name Graph&Port operations
	//@{
	void LinkClear();										///< Clear all links to/from this device
	Port *GetPort(int index, int io);						///< Return a pointer to the given port
	int  GetNextPort(int io);								///< Returns the next unconnected port index of a type, or -1 if no unconnected ports of that type exist
	bool CheckPort(int slot, int io);						///< Is the specified port connected?
	int  GetPortStyle(int slot, int io);					///< Return if the port is flagged special or not
	bool FindPort(const Port *port, int *index, int *io);	///< Look for the specified port and return its slot number and IO type on THIS device, if it exists
	int  GetLinkedPort(Port *linked_port, int io = -1);		///< Given a port on a DIFFERENT device, return the slot number of the connected port on THIS device

	/// Safely retrieve one of a device's input ports
	InputPort *in(int slot) { if (slot < (int) inputs.size()) return inputs[slot]; else return NULL;};
	/// Safely retrieve one of a device's output ports
	OutputPort *out(int slot) { if (slot < (int) outputs.size()) return outputs[slot]; else return NULL; };
	/// Safely retrieve one of a device's parameter ports
	InputPort *parmin(int slot) { if (slot < (int) params_input.size()) return params_input[slot]; else return NULL; };
	/// Return the mask port
	InputPort *mask() { return mask_input; };	
	/// Return the coordinate port
	InputPort *coord_port() { return coord_input; };	

	int  highest_input();	/// Retrieve index of the highest connected INPUT. -1 == no connections
	int  highest_output();	/// Retrieve index of the highest connected OUTPUT
	int  active_inputs();	/// number of connected inputs
	int  active_outputs();	/// number of connected outputs
	int  active_param_inputs();	/// number of connected param inputs
	/// Number of inputs
	int	 max_inputs() const { return (int) inputs.size(); };
	/// Number of outputs
	int  max_outputs() const { return (int) outputs.size(); };
	/// Number of parameter slots
	int  max_parm() const	{ return (int) params_input.size(); };
	//@}

/// @name Device functionality
/// These are the heart of the device, and control both the actual effect and the interaction with the Device.
	//@{
	virtual bool RunUI();				///< Run the user interface for this part. Over-ride to show a custom dlg. Returns true if the device needs to be rebuilt.
	bool		 RunNameDialog();		///< Run the user interface to set a new name for this part
	virtual void RunHelpDialog();		///< Run the user interface to display a dialog with info about the device

	/// If you provide a BuildContext that contains areapacket-type (proxy or real) packets, calling this will calculate the correct resolution for all inputs and outputs of this device,
	/// and fill all inputs, outputs, and mask input with AreaProxy packets containing the relevant resolutions.
	/// This function cannot be overridden, but you can change how your device calculates its output sizes by overloading the protected function doCreateOutputProxy().
	/// This should only be necessary if your device has mixed-or-unusual output resolutions.
	void CalculateProxySizes(BuildContext proxycontext);

	virtual ACTIVATE_RESULT ActivateSequence(BuildContext &context);	///< Wraps call to PreAct, ActOnNeed, and PostAct
	virtual ACTIVATE_RESULT ActivateOnNeed(BuildContext &context);		///< Calls Activate only if changed. Generally don't override this
protected:
	virtual bool Activate(BuildContext &context) = 0;			///< THE ALL IMPORTANT Activate call. Actually PERFORM the device action!
public:
	//@}

	virtual bool RunAuxAction(int index);		///< Run extra actions defined by this sprocket. Returns TRUE if framework should re-calculate after action
	virtual int  getAuxActionSize();			///< The number of actions defined
	virtual char *getAuxActionName(int index);	///< The name of the extra action


	DEV_DATA &GetLifeVars() { return lifeptrs; };
	virtual bool Randomize(int newseed = 0);		///< Randomize this device, if applicable.

	void Preview();			///< Call parent world to have this device rebuilt in the preview
	virtual void Falsify();	///< Inform parent world that we have been falsified

/// @name Parameter Controls. The reference versions are allowed to throw exceptions
/// @throws WMException
	//@{
	int		GetI( const char *name);
	float	GetF( const char *name);
	bool	GetB(const char *name);

	int		&GetIRef( const char *name);		///< Get a reference to the given integer parameter
	float	&GetFRef( const char *name);		///< Get a reference to the given float parameter
	bool	&GetBRef(const char *name);			///< Get a reference to the given boolean parameter
	dword   &GetCRef(const char *name);			///< Get a reference to the given RGB Color parameter

	Parameter &GetParm(const char*name);		///< Get a reference to the given parameter

	ParameterSet &GetParams() { return params; };
	void SetParams(ParameterSet &other) { params = other; }; //params.setOwner(this); };
	PacketPointer RetrieveParamPacket( int slot );

	/// Enumerate the different versions this device supports
	virtual int numVersions();
	virtual char const* getVersionName(int ver);
	virtual int getVersion() const;
	/// Change to a new parameter version set, saving the current settings into the version controller and making a new version current.
	virtual void ChangeVersion(int versionnum);
	/// If your device was previously unversioned, during Device Load you must use SetInitialVersion() before calling the superclass Load()
	/// to ensure that the parameters are being loaded into the correct version slot.
	virtual void SetInitialVersion(int versionnum);
	/// If introducing versioning to a device that was previously unversioned, in your constructor you should call SetLegacyVersion to instruct
	/// the device to associate the previously saved parameters with one of the newly introduced versions
	virtual void SetLegacyVersion(int versionnum);
	/// Create a version from the given parameterset
	virtual int AddVersion(char const*name, ParameterSet const&params);

	//@}


	/// Notification/Event enums
	// World events that a device may or may not depend on. Right now just a few of these...
	// used by isUpdatedOn to control if the device is reset or not on this event
	enum DEVICE_EVENT { DEV_EVENT_LIGHTING = 1, DEV_EVENT_WATER = 2, DEV_EVENT_COLORTABLE = 4, DEV_EVENT_RENDER_EXTENTS = 8 };

	/// @name Notification Functions
	/// Provide a means for a Device to be told about certain events happening in the world. This can be useful both internally (things like
	/// morphing output ports, etc) and externally (synchronizing with another application)
	///
	//@{

	/// Notification that this device has been added to a world. This will occur after a file is fully loaded, or immediately after a device
	/// is added to an already-existing world. The loaded parameter indicates if this was from a file load or simply a part added.
	virtual void notifyAdded(bool loaded);

	/// Notification that this device has been removed from a world and is about to be destroyed
	virtual void notifyRemoved();

	/// Notification that this device's world is about to be closed. This occurs when a file is closed or when World Machine exits
	virtual void notifyWorldClosing();

	/// Notification that the user has changed the version of this device. The framework will autoamtically change the parameters
	/// for you in the internalChangeVersion() function, but if you need to do other things (add or remove ports, for example),
	/// then do them here.
	virtual void notifyChangedVersion(int oldVersion, int newVersion);

	/// Notification when a wire is attached or removed to a port for this device.
	/// You will receive the notification AFTER the event in question has occured; this may be useful for doing things like morphing port type, etc.
	virtual void notifyPortConnected(int slot, int io, bool is_connecting);

	/// Notification that this device has been falsified, and its contents have been removed/are no longer valid.
	virtual void notifyFalsify();

	// Notification that some world settings have been changed. The specific type of change is specified by the event parameter
	virtual void notifyWorldChanged(DEVICE_EVENT event, VAL value);

	/// Notification sent just before a final-mode build begins. the tiled boolean indicates if it is a tiled or regular build
	/// Note that you will receive this before any devices begin building
	virtual void notifyBuildStarting(bool tiled);

	/// Notification sent when a build has completed. the tiled boolean indicates if it is a tiled or regular build
	/// You will receive this when all devices that are queued to build are complete.
	/// Note that for a Tiled Build, this means that all tiled have been blended as well.
	virtual void notifyBuildEnd(bool tiled);

	//@}

	/// @name Context Sensitive Queries
	/// These are convenience wrappers around the objects world state tat is actually stored in the passed-in buildContext
	bool isBuilt(PSTATE state);
	bool isBuilt(BuildContext &context);
	/// Is this device asleep (signal'd to not build because of child caching
	bool		 isAsleep() { return false; }; //{ return (processed & DEV_SLEEPING) != 0; };	

	/// @name Interrogatives
	/// Dynamic state of the device
	//@{
	/// device is ghosted if it doesn't have a current owner world. Usually used within the UI to show a device ready to be placed
	bool		 isGhosted() { return !worldptr; };	
	/// Is this device toggled to disable mode?
	bool		 isEnabled() { return enable; };	
	/// Is this device bypassed?
	bool		 isBypassed() { return bypass; };	
	//@}
	/// enable/disable this device. A disabled device does not produce any output
	void		setEnable(bool _en) { enable = _en; };	
	/// set device to be bypassed or not. A bypassed device simply passes its inputs to its outputs with no further processing
	void		setBypass(bool _by) { bypass = _by; };

	/// If this device should be considered "canonical"; that is: is this device part of the actual deviceworld, or does it
	/// reside within a copy of the world. If your device is communicating with the outside world, you should only do so if you are the
	/// canonical device -- otherwise copies from layout view, undo/redo, etc will get in the way.
	bool		isCanonical();

	/// @name Intrinsic Interrogatives
	/// These reflect the capabilities of the device class. Your class should implement these if your answers are different from the defaults.
	//@{
	/// Device returns true if it needs to be recreated when the given world parameter changes
	virtual bool isUpdatedOn(int) { return false; }; 
	/// If this device requires the registered version or not.
	virtual bool isRegistered() { return false; };	
	/// This device has a sensible output for viewing? 
	virtual bool isViewable() { return true; };		
	/// Is this device
	virtual bool isBypassable() { return true; };

	/// Should this device be un-disablable & undeletable in the workview?
	virtual bool isPermanent() { return false; };	
	/// Allow presets for this device in the generic dialog manager?
	virtual bool isPresetCapable() { return true; };
	/// Allow the user to assign a resolution override to this device?
	virtual bool isOverrideCapable() { return true; };
	/// If this device is compatible with explorer mode or not
	virtual bool isExplorable() { return true; };   
	/// Should this device have a mask input?
	virtual bool isMaskable() { return false; };
	/// Should the framework automatically use the mask input to blend the effect? If isAutoMasking() is false, the device must handle its own masking when isMaskable() is true.
	virtual bool isAutoMasking() { return true; };
	/// Does this device use the origin value?
	virtual bool hasOrigin() { return false; };		
	/// Does this device allow local space?
	virtual bool allowSpatialOverride() { return false; }; 
	/// Is this device a building checkpoint
	virtual bool isCheckPoint() { return false; };
	/// Is this device (or device version) deprecated? User will be warned if using a deprecated device.
	virtual bool isDeprecatedVersion();
	//@}

	/// @name Hinting
	//@{
	virtual int getSpeedClass() { return 3; };		// Provide a rough estimate of processing speed of this device, relative to perlin noise (3). Higher is slower
	virtual int  estimateMemoryUsage();				// Provide the framework with an estimate of the memory required by this device given current settings, in KB
	/// Return the number of bytes of extra storage used by this device; defaults to zero. Override if your device will keep substantial (>1MB) per-device data
	virtual size_t getHeapUsage() { return 0; };

	void setMaskOverride(int value = 0);	// Temporarily override how this device outputs data
	int  getMaskOverride() { return mask_override; };
	//@}

	/** @name Overrides
		Overrides allow you to control the resolution or spatial type (world/localspace) independent of the worlds' settings. Devices have their own override, but in addition
		if they belong to a group the group may supply an override as well.
	 */
	//@{

	/// If this device operates in world or local space. Note that this is specifically the DEVICE's override, not the final override state that is influenced by any groups
	/// the device is a member of.
	PacketContext::SPACETYPE getSpatialOverride() const { return devicecontext.getSpatialType(); };	
	/// Set worldspace behavior of the device
	void setSpatialOverride(PacketContext::SPACETYPE t) { devicecontext.setSpatialType(t); };

	/// Does this device have an active override enabled?
	bool hasResolutionOverride() const;
	/// Does this device have an active spatial override enabled?
	bool hasSpatialOverride() const;

	/// Return a composite override consisting of the composited device & group override for the current device. If none is active, it will always return the devices'.
	WorldMachine::ContextOverride getComposedOverride() const;

	/// Retrieve specifically this device's context override
	WorldMachine::ContextOverride &getDeviceContextOverride() { return devicecontext; };

	/// Select one of this device's output ports as an override of the default port.
	/// If supplied with -1 (the default), the device's built-in default output is used.
	void selectDefaultOutput(int port = -1);
	int	 getSelectedDefaultOutput() const { return default_output_port; };

	//@}

protected:
	virtual void internalChangeVersion(int versionnum);

	virtual void doClone(Device *other);									/// Copy data from other to this

	virtual void doCreateOutputProxy(BuildContext &proxycontext);			// Process proxy packets for this device to estimate resolution across the network

	// override to control how we mask against inputs. Default does no masking.
	virtual bool doMaskCreateNewCopies(std::vector<PacketPointer> &results, BuildContext &context);
	/// override to do your own masking
	virtual void doMaskOperation(PacketPointer _result, PacketPointer _original, PacketPointer _mask);	// Actual blending operation

/// @name Data input/output/activation functions
/// Note that in our current model these are simply convenience wrappers to the appropriate BuildContext functions
	//@{
	void StoreData(PacketPointer const&dat, int slot, BuildContext &context);			// Stores a completed HF in our output buffer
	PacketPointer RetrieveInput(int slot, BuildContext &context);				// Retrieve main inputs data, and TRANSFERS ownership
	HFPointer	  RetrieveMask(BuildContext &context);
	PacketPointer RetrieveTransform(BuildContext &context);
	//@}

	bool Bypass(BuildContext &context);						///< Perform a bypass of this device, if possible

	DEV_DATA	lifeptrs;			///< make&destroy pointers for this device
	DeviceWorld *worldptr;			///< pointer to this device's container
	Coord loc;						///< Coordinate of this device in world space

	bool enable;					///< If this device is disabled from activation or not.
	bool bypass;					///< If this device should be bypassed
	int  mask_override;				///< 0 = normal masking
									///< 1 = override mask on
									///< 2 = override mask off

	int default_output_port;		///< Contains the port number of the assigned default output, or -1 if default

	void SetLinks(int in, int out, WorldMachine::PacketType packet_type = WorldMachine::Packet::hfield, int common_flags = 0);	 // Setup this device's link vector
	void ChangeLinks(int in, int out, WorldMachine::PacketType packet_type = WorldMachine::Packet::hfield, int common_flags = 0);// Change the # of links without destroying existing ones. Used by devices
	void ChangeParamLinks(int in);
	void ChangeLinkIn(int in, WorldMachine::PacketTypeMany packet_type = WorldMachine::Packet::hfield, int common_flags = 0);		// for on-the-fly modifications of the # of links
	void ChangeLinkOut(int out, WorldMachine::PacketType packet_type = WorldMachine::Packet::hfield, int common_flags = 0);
	void MorphOutput(int slot, WorldMachine::PacketType packet_type);					///< Change the type of a single link, breaking any existing links that are no longer of the right type

	virtual void AddPort(int flag, int io, WorldMachine::PacketType packet_type = WorldMachine::Packet::hfield, int index = -1);
	virtual void RemovePort(int io, int index );


	virtual void AddParam(Parameter p);				///< Adds a parameter to paramset, and also creates an override port for it
	virtual void RemoveParam(int index);			///< Remove a parameter, delete its override port
	WorldMachine::DeviceVersioning	versioncontrol;	///< Mechanism to allow different revisions of a device to have different parameters.
	ParameterSet			params;					///< Set of user controllable, managed-parameters

	std::vector<InputPort*>		inputs;			///< Input ports
	std::vector<OutputPort*>	outputs;		///< Output ports 
	std::vector<InputPort*>		params_input;	///< Parameter-override hookups

	InputPort *mask_input;				///< Built-in mask input
	InputPort *coord_input;				///< Built-in coordinate input

	char name[DEV_NAME_LEN];		///< name of this device

	WorldMachine::ContextOverride devicecontext;
	bool helperLoadFunc(std::istream &in, int &ver);
	void helperSaveFunc(std::ostream &out, int ver);
private:
	CoordF origin;					///< the device's origin.
	DEVID device_uid;			///< Device unique identifier. NOT globally unique - guaranteed unique only within the current world file.

	int legacy_version; // this is set by SetLegacyVersion and is not load/saved. It instructs the device how to handle new versions appearing

	void ChangedLinks();
	virtual void ResetParams();							///< Deletes all params and param ports

	int doStoreDistribute(BuildContext &context);
	void fillAllInputs( BuildContext & context );
	PacketPointer reachThroughInput(BuildContext &context, Port *port);

	bool ProcessInputs( BuildContext & context );
	void validateInputPackets( DeviceResults &results, InputValidationStruc &out_validate );
	bool processDeviceContextOverride( BuildContext &context );

	/// visit all inputs and mask input.
	/// Functor type:
	///   void functor(InputPort *p, int slot, int io)
	template < class Functor >
	void visitAllInputs(Functor f) {
		for (int slot = 0; slot < max_inputs(); slot++) {
			f( in(slot), slot, Device::INPUT);
		}
		f( mask_input, 0, Device::MASK);
	}
};

/** Simple, serializable identification of a port. World Machine uses PortInfoStruc instead of direct pointers to ports frequently.
 This is in the Device.h header instead of the Port.h for dependency reasons.
*/
struct PortInfoStruc {
	PortInfoStruc() { device = nullptr; port = -1; io = Device::NULLPORT; };
	PortInfoStruc(Device *dev, int _port, Device::DEVICEPORT _io) { setup(dev, _port, _io); };
	// Construct by looking up port and filling out info
	PortInfoStruc(Port *p);

	void setup(Device *dev, int _port, Device::DEVICEPORT _io) { device = dev; port = _port; io = _io; };
	bool valid() { return device && io != Device::NULLPORT && port != Device::NULLPORT; };

	/// Retrieve a port from the given port information, or nullptr if cannot.
	Port *getPort();

	Device *device;
	int port;
	Device::DEVICEPORT io;
};

#endif
