#pragma once

#include <vector>
#include "WMAllocator.h"
#include "Parameter.h"

namespace WorldMachine {

class DeviceVersioning
{
public:
	DeviceVersioning(ParameterSet &owner_);
	~DeviceVersioning(void);
	DeviceVersioning &operator=(DeviceVersioning const&other);

	/// Version change function is called whenever the version changes.
	/// The purpose is to allow your device to change other things that differ between versions, such as add new input ports, or import parameters from an old version on first use, etc.
	/// This is called after the current params are saved back to the version store but before the new version store is set to the device.
	/// Return false from your callback if you want to force-cancel the version change.

	/// Parameters:
	///  (int) version changing from
	///  (int) version changing to
	/// Return: true to allow change to proceed, or false to cancel the version change
	typedef std::function<bool(int, int)> ChangeFunc;
	void setVerChangeFunc(ChangeFunc const&changeHandlers);

	void Save(std::ostream &out);
	void Load(std::istream &in);

	int numVersions() { return (int) versions.size(); };

	/// Return the version number currently active. -1 means the current device is unversioned.
	int getCurVersion() const;

	/// Return true if using a non-current version. This is any version below the last version added to the versioning class.
	bool isCurrentVersionDeprecated();

	/// Save the currently set owner parameters to the version store
	void saveCurrentVersion();

	/// Change the active version -- this will repopulate the linked owner parameter set to one of our stored ones
	bool changeVersion(int version);


	/// Add a version to the versioning bank
	int addVersion(const char* name, ParameterSet const&params);

	/// Return the name of the given version
	const char*getVersionName(int version);

	/// Retrieve a reference to the parameter set associated with the given version.
	/// Will throw if that version doesn't exist.
	ParameterSet &getVersion(int version);

	/// If the version queried has been previously saved to before
	bool isPreviouslySaved(int version);

	/// This is a special case command, used only when versioning has been added to a device that previously did not have a version.
	/// It forces the version store to forget the current device version.
	void forceVersion( int version);

	/// Resets the "used/visited" tag for all versions. Useful mostly for Device::SetInitialVersion
	void resetVisited();

private:
	DeviceVersioning(DeviceVersioning const&other);
	int current_version;
	struct VersionSet {
		bool used;
		int version_number;
		char name[PARAM_NAME_SIZE];
		ParameterSet params;
	};
	std::vector<VersionSet> versions;
	ParameterSet &owner;
	ChangeFunc changefunc;
};

}
