#pragma once
#include "PacketVisitor.h"
#include "ThreadHelper.h"

namespace WorldMachine {

/**
 This specialization is a drop-in replacement for the packet visitor class that spawns worker threads and visits pixels in parallel.

 The only contract that you must follow is that each invocation of your visitor function must only _write to the current element_.
 You may read from any element (thus kernel-type processes are OK), but only the current element can be written to.

 Of course, in the example of a blur kernel, you'd still want to write to a different heightfield than the one you're reading from, to prevent
 contamination of source and destination..
 
*/
template < class PacketType >
class ParallelPacketVisitor : public PacketVisitor<PacketType> {
public:
	// If you need to create with more than 4 packets, call add() yourself to insert the extras.
	ParallelPacketVisitor(PacketType &first) : PacketVisitor(first) { allow_parallel = true; };
	ParallelPacketVisitor(PacketType &first, PacketType &second) : PacketVisitor(first, second) {  allow_parallel = true; };
	ParallelPacketVisitor(PacketType &first, PacketType &second, PacketType &third)  : PacketVisitor(first, second, third) {  allow_parallel = true; };
	ParallelPacketVisitor(PacketType &first, PacketType &second, PacketType &third, PacketType &fourth)  : PacketVisitor(first, second, third, fourth) {  allow_parallel = true; };

	void setParallel(bool allow_parallel_) { allow_parallel = allow_parallel_; };

	/// Creates and attaches simple callback function for you based upon the current build context and device.
	/// Also will respect the build context's device threading command
	void setStandardCallback( Device* dev, BuildContext & context ) 
	{
		setParallel(context.allow_device_threading);
		PacketVisitor<PacketType>::setStandardCallback(dev, context);
	}


	template < typename VisitFunc >
	bool visitRegion( Rect4 region, VisitFunc visitor ) {
		RectF area = areacontext.getRenderExtent();
		Vector2D scaling(area.width / (double(areacontext.w())-1.0), area.height / (double(areacontext.h())-1.0));

		size_t packetsize = packets.size();

		Location baseloc;
		baseloc.wloc = CoordF(area.x, area.y);
		baseloc.loc = Coord(region.x0,region.y0);
		baseloc.index = 0;
		baseloc.itr.resize(packetsize);
		for (size_t i = 0; i < packetsize; ++i) {
			baseloc.itr[i] = packets[i]->GetDataPtr();
		}

		int w = areacontext.w();
		int h = areacontext.h();

		int numworkitems = region.y1 - region.y0;

		Rect4 creg(0, 0, w, h);
		clipRegion(region, creg);

		ThreadHelper *helper_ptr;

		auto scanlineProcessor = [&](int workitem) -> bool {
			Location loc = baseloc;
			loc.wloc.x  = area.x + float(double(region.x0) * scaling.y);
			loc.wloc.y = area.y + float(double(region.y0 + workitem) * scaling.y);
			loc.loc.x = region.x0;
			loc.loc.y = region.y0 + workitem;
			loc.index = areacontext.getIndexFromGrid(loc.loc);
			for (size_t i = 0; i < packetsize; i++) {
				loc.itr[i] = baseloc.itr[i] + loc.index;
			}

			for (; loc.loc.x < region.x1;) {
				visitor(loc); // invoke

				// update all locs and ptrs
				loc.wloc.x = float(double(loc.wloc.x) + scaling.x);
				++loc.loc.x;
				++loc.index;
				for (size_t i = 0; i < packetsize; i++) {
					++loc.itr[i];
				}
			}

			// call the progress functor every scanline, and break if it indicates to abort.
			if (progress_callback && !progress_callback(helper_ptr->numProcessed(), numworkitems))
				return false;

			return true;
		};

		ThreadHelperWrapper<decltype(scanlineProcessor)> thread_dispatch(scanlineProcessor);
		thread_dispatch.setWorkItems(numworkitems);
		thread_dispatch.setThreading(allow_parallel);
		// Do the traversal with manual pointer arithmetic while updating the world and HF-space coordinates appropriately
		helper_ptr = &thread_dispatch;
		thread_dispatch.PerformWork();
		return true;
	};

private:
	bool allow_parallel;
};

}
